/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * Authors: Andreas Fenkart, <andreas.fenkart@dev.digitalstrom.org>
 *
 * digitalSTROM ds485-base library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-base library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-base library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "log.h"

#include <ctype.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <syslog.h>

/** log entry size excluding timestamp, incl. newline log source description */
#define MAX_LOGLINE 256

static struct {
    int debugLevel;
    int use_syslog;
    char *log_prefix;
} logger;

/* protects global buffer in my_log3000 */
static pthread_mutex_t reportMutex = PTHREAD_MUTEX_INITIALIZER;

void init_logger(char *prefix, int level, int use_syslog) {
    logger.log_prefix = strdup(prefix);
    logger.debugLevel = level;
    if (use_syslog) {
        enable_syslog(use_syslog);
    }
}

void cleanup_logger()
{
    free(logger.log_prefix);
    enable_syslog(false);
}

void set_log_level(int level)
{
    logger.debugLevel = level;
}

void enable_syslog(int enable)
{
    if (enable) {
        logger.use_syslog = true;
        openlog(logger.log_prefix, LOG_PID, LOG_USER);
    } else {
        logger.use_syslog = false;
        closelog();
    }
}

static void my_log3000(const char *buf)
{
    char tsbuf[30], buf2[MAX_LOGLINE + 30];
    const char* sp;
    struct timeval t;

    gettimeofday(&t, NULL);
    strftime(tsbuf, sizeof(tsbuf), "%Y-%m-%d %H:%M:%S", localtime(&t.tv_sec));
    (void)snprintf(buf2, sizeof(buf2), "[%s.%03d] ", tsbuf, (int)t.tv_usec / 1000);

    for (sp = buf; *sp != '\0'; sp++) {
        /* TODO buffer overlow if user can control output */
        /* if hex values maps to '\n' only print it if at the end of buffer */
        if (isprint((int)*sp) || (isspace((int)*sp) && (sp[1]=='\0' || sp[2]=='\0')))
            (void)snprintf(buf2+strlen(buf2), 2, "%c", *sp);
        else
            (void)snprintf(buf2+strlen(buf2), 6, "\\%02hhx", *sp);
    }

    (void)fputs(buf2, stderr);
}

void report(int errlevel, const char *fmt, ... )
{
    va_list ap;
    va_start(ap, fmt) ;
    vreport2(errlevel, NULL, fmt, ap);
    va_end(ap);
}

void report2(int errlevel, const char *id, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt) ;
    vreport2(errlevel, id, fmt, ap);
    va_end(ap);
}

static void do_emit(int level, const char *buf)
{
    pthread_mutex_lock(&reportMutex);

    if (logger.use_syslog) {
        syslog(level, "%s", buf);
    } else {
        my_log3000(buf);
    }

    pthread_mutex_unlock(&reportMutex);
}

static char last_msg[MAX_LOGLINE];
static int last_level;
static int repetition;
void vreport2(int errlevel, const char *id, const char *fmt, va_list ap)
{
    char buf[MAX_LOGLINE];

    if (errlevel > logger.debugLevel) {
        return;
    }

    if (id) {
        snprintf(buf, sizeof(buf), "%s: <%s> ", logger.log_prefix, id);
    } else {
        snprintf(buf, sizeof(buf), "%s: ", logger.log_prefix);
    }

    vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), fmt, ap);

    if (strncmp(buf, last_msg, strlen(buf)) == 0) {
      repetition++;
      return;
    }

    if (repetition) {
      char tmp[80];
      snprintf(tmp, sizeof(tmp), "%s: ... message repeated %d times (next line)\n",
               logger.log_prefix, repetition);
      do_emit(last_level, tmp);
      do_emit(last_level, last_msg);
    }

    repetition = 0;
    memcpy(last_msg, buf, sizeof(last_msg));
    last_level = errlevel;
    do_emit(errlevel, buf);
}

#if 0
int main()
{
    const char buf[] = "last 'x' char will not be printed x\n";
    const char buf2[] = { 0x48, 0x61, 0x14, 0x34, 0x12, 0x45, 0x0a };
    const char buf3[] = "newline 1\n newline2\n";

    my_log3000(buf);
    my_log3000(buf2);
    my_log3000(buf3);
}
#endif
